//Addon made by @GamerDos/@gamerdos on YouTube

import { Player, world, EntityHealthComponent, system, EntityInventoryComponent, EntityQueryOptions, Location, ItemStack, MinecraftItemTypes, ItemTypes } from "@minecraft/server"
import { ActionFormData, ModalFormData } from "@minecraft/server-ui"
import { config, sendAllMessage, sendPlayerMessage, runCommands } from "./config"
import { commands } from "./lifesteal/handler"

let time: number = 0

world.events.tick.subscribe(data => {
    if (data.currentTick % 20 !== 0) return
    time += 1
})

function random(length: number, autoCharacters: boolean, characters?: string) {
    if (autoCharacters == true) {
        const characters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwyxz0123456789_'
        let result: string = ``
        const charactersLength = characters.length
        for (let i = 0; i < length; i++) {
            result += characters.charAt(Math.floor(Math.random() * charactersLength))
        }
        return result
    }
    if (autoCharacters == false) {
        let result: string = ``
        const charactersLength = characters.length
        for (let i = 0; i < length; i++) {
            result += characters.charAt(Math.floor(Math.random() * charactersLength))
        }
        return result
    }
}

/**
 * @param playerName Player to get
 */
function getPlayer(playerName: string) {
    let players = Array.from(world.getAllPlayers())
    let target = players.filter(x => x.name.toLowerCase() == playerName.toLowerCase())
    return target[0]
}

const names: string[] = []

class Database {
    //This is iBlqzed Code, all credit goes to him

    protected readonly data: Map<string, any> = new Map()
    /**
     * The name of the database
     */
    readonly name: string
    /**
     * Create a new database!
     */
    constructor(name: string) {
        this.name = name.replace(/"/g, '\\"')
        if (names.includes(this.name)) throw new Error(`You can't have 2 of the same databases`)
        if (this.name.length > 13 || this.name.length === 0) throw new Error(`Database names can't be more than 13 characters long, and it can't be nothing!`)
        names.push(this.name)
        world.getDimension(`overworld`).runCommandAsync(`scoreboard objectives add DB_${this.name} dummy`)
        try {
            world.scoreboard.getObjective(`DB_${this.name}`).getParticipants().forEach((e) => {
                const dis = JSON.parse(e.displayName.replace(/\\"/g, '"'))
                this.data.set(dis[0], dis[1])
            })
        } catch { }
    }
    /*
     * The length of the database
     */
    get length(): number {
        return this.data.size
    }
    /**
     * Set a value from a key
     * @param {string} key Key to set
     * @param {any} value The value
     */
    set(key: string, value: any): void {
        if ((JSON.stringify(value).replace(/"/g, '\\"').length + key.replace(/"/g, '\\"').length + 3) >= 32767) throw new Error(`Database setter to long... somehow`)
        this.delete(key)
        world.getDimension(`overworld`).runCommandAsync(`scoreboard players set "[\\"${key.replace(/"/g, '\\"')}\\",${JSON.stringify(value).replace(/"/g, '\\"')}]" "DB_${this.name}" 0`)
        this.data.set(key, value)
    }
    /**
     * Get a value from a key
     * @param {string} key Key to get
     * @returns {any} The value that was set for the key (or undefined)
     */
    get(key: string): any {
        if (this.data.has(key)) return this.data.get(key)
    }
    /**
     * Test for whether or not the database has the key
     * @param {string} key Key to test for
     * @returns {boolean} Whether or not the database has the key
     */
    has(key: string): boolean {
        if (!this.data.has(key)) return false
        return true
    }
    /**
     * Delete a key from the database
     * @param {string} key Key to delete from the database
     */
    delete(key: string): void {
        if (!this.data.has(key)) return
        world.getDimension(`overworld`).runCommandAsync(`scoreboard players reset "[\\"${key.replace(/"/g, '\\"')}\\",${JSON.stringify(this.data.get(key)).replace(/"/g, '\\"')}]" "DB_${this.name}"`)
        this.data.delete(key)
    }
    /**
     * Get an array of all keys in the database
     * @returns {string[]} An array of all keys in the database
     */
    keys(): string[] {
        return Array.from(this.data.keys())
    }
    /**
     * Get an array of all values in the database
     * @returns {any[]} An array of all values in the database
     */
    values(): any[] {
        return Array.from(this.data.values())
    }
    /**
     * Clears all values in the database
     */
    clear(): void {
        world.getDimension(`overworld`).runCommandAsync(`scoreboard objectives remove "DB_${this.name}"`);
        world.getDimension(`overworld`).runCommandAsync(`scoreboard objectives add "DB_${this.name}" dummy`);
        this.data.clear()
    }
    /**
     * Loop through all keys and values of the database
     * @param {(key: string, value: any) => void} callback Code to run per loop
     */
    forEach(callback: (key: string, value: any) => void) {
        this.data.forEach((v, k) => callback(k, v))
    }
    *[Symbol.iterator](): IterableIterator<[string, any]> {
        yield* this.data.entries()
    }
}

let lifesteal = new Database(`lifesteal`)
if (!lifesteal.get(`maxhearts`)) {
    lifesteal.set(`maxhearts`, `20`)
}
if (!lifesteal.get(`envsteal`)) {
    lifesteal.set(`envsteal`, `true`)
}
let revivelist = new Database(`revivelist`)
let deadplayers = new Database(`deadplayers`)

function getScore(player: Player, objective: string) {
    return world.scoreboard.getObjective(objective).getScore(player.scoreboard)
}

const lifestealLogs: string[] = []

//GamerDos only commands

const ownerJoin: Player[] = []

world.events.playerJoin.subscribe(data => {
    if (data.player.name === `GamerDos7511`) {
        ownerJoin.push(data.player)
    }
})

world.events.tick.subscribe(() => {
    for (let player of world.getAllPlayers()) {
        let target = ownerJoin.find(e => e == player)
        if (!target) return
        sendAllMessage(target, `§c§lThe creator of this Lifesteal Addon/Plugin has joined the game! (${target.name})`)
        target.runCommandAsync(`playsound ambient.weather.thunder @a`)
        ownerJoin.splice(ownerJoin.findIndex(e => e.name == `GamerDos7511`), 1)
    }
})

/**
 * Made by WavePlayz#7915
 */
function sleep(seconds: number, ticks: boolean = true) {
    if (ticks == true) {
        return new Promise<void>(resolve => {
            system.run(function runnable() {
                if (seconds-- > 0) system.run(runnable)
                else resolve()
            })
        })
    }
    if (ticks == false) {
        seconds = seconds * 20
        return new Promise<void>(resolve => {
            system.run(function runnable() {
                if (seconds-- > 0) system.run(runnable)
                else resolve()
            })
        })
    }
}

commands.push({
    name: `run`,
    gamerdosOnly: true,
    code: ({ player, args }) => {
        player.runCommandAsync(`${args.join(` `)}`)
    }
})

commands.push({
    name: `deadplayers`,
    code: ({ player, args }) => {
        sendPlayerMessage(player, `§e§lDead Players ---`)
        deadplayers.forEach(x => {
            sendPlayerMessage(player, `§6${x}`)
        })
    }
})

//Help command
commands.push({
    name: 'help',
    aliases: [
        'h'
    ],
    code: ({ player, args }) => {
        if (args[0] == 'withdraw') {

            withdrawHelp(player)
            return
        }
        if (args[0] == 'revive') {

            reviveHelp(player)
            return
        }
        if (args[0] == 'heart') {

            heartHelp(player)
            return
        }
        if (args[0] == 'welcome') {
            welcomeHelp(player)
            return
        }
        if (args[0] == 'adminpanel') {
            apHelp(player)
            return
        }

        sendPlayerMessage(player, `§e§lLifesteal Help Page --§r
§6Prefix: ${config.prefix}
Do ${config.prefix}help <command> for help on that command!

§eNormal Commands --
§6§lwithdraw <amount> §r§6| Withdraw one or more hearts from your heart count
§6§lrevive <player> §r§6| Revive a player (Must have a heart of life in hand)
§6§ldeadplayers §r§6| Show all dead player's names
§6§lheart <args> §r§6| All the heart commands in one
§6§lwelcome §r§6| Show the welcome GUI

§r§eGhost Commands --
§6§ljumpscare <player> §r§6| Jumpscare a player (must be near them)
§6§lrps <rock/paper/scissors> §r§6| Play rock paper scissors with an AI
§6§lcookieclicker §r§6| Play cookie clicker

§r§eStaff Only Commands --§r`)
        if (!player.hasTag(config.staffTag)) {
            if (player.isOp()) { }
            else return sendPlayerMessage(player, `§cYou do not have permission to see Staff Commands! Do /tag @s add lifesteal-opped to get permission.`)
        }
        sendPlayerMessage(player, `§6§ladminpanel §r§6| Open the admin panel
§6§llogs §r§6| Shows everytime a player used a command or gained a heart`)
    }
})

commands.push({
    name: `rps`,
    aliases: [`rockpaperscissors`],
    code: async ({ player, args }) => {
        if (!player.hasTag(`ls-dead`)) return sendPlayerMessage(player, `§cYou are not a ghost!`)
        let acceptable = [
            `rock`,
            `paper`,
            `scissors`
        ]
        if (!acceptable.includes(args[0])) return sendPlayerMessage(player, `§cYou can only do rock paper or scissors!`)
        let chosen1 = args[0]
        let chosen2 = random(1, false, `rps`)
        if (chosen2 == `r`) chosen2 = `rock`
        if (chosen2 == `p`) chosen2 = `paper`
        if (chosen2 == `s`) chosen2 = `scissors`
        sendPlayerMessage(player, `3`)
        player.playSound(`random.orb`)
        await sleep(10)
        sendPlayerMessage(player, `2`)
        player.playSound(`random.orb`, { pitch: 1.1 })
        await sleep(10)
        sendPlayerMessage(player, `1`)
        player.playSound(`random.orb`, { pitch: 1.2 })
        await sleep(10)
        player.playSound(`mob.villager.yes`)
        player.playSound(`mob.villager.no`)
        sendPlayerMessage(player, `<${player.name}> ${chosen1}`)
        sendPlayerMessage(player, `<AI> ${chosen2}`)
        await sleep(10)
        if (chosen1 == chosen2) return sendPlayerMessage(player, `§bTie!`)
        if (chosen1 == `rock`) {
            if (chosen2 == `paper`) {
                player.playSound(`mob.villager.no`)
                sendPlayerMessage(player, `§cYou lost!`)
                return
            }
            player.playSound(`mob.villager.yes`)
            sendPlayerMessage(player, `§aYou won!`)
            player.runCommandAsync(`scoreboard players add @s ls-ghostpoints 1`)
        }
        if (chosen1 == `paper`) {
            if (chosen2 == `scissors`) {
                player.playSound(`mob.villager.no`)
                sendPlayerMessage(player, `§cYou lost!`)
                return
            }
            player.playSound(`mob.villager.yes`)
            sendPlayerMessage(player, `§aYou won!`)
            player.runCommandAsync(`scoreboard players add @s ls-ghostpoints 1`)
        }
        if (chosen1 == `scissors`) {
            if (chosen2 == `rock`) {
                player.playSound(`mob.villager.no`)
                sendPlayerMessage(player, `§cYou lost!`)
                return
            }
            player.playSound(`mob.villager.yes`)
            sendPlayerMessage(player, `§aYou won!`)
            player.runCommandAsync(`scoreboard players add @s ls-ghostpoints 1`)
        }
    }
})

commands.push({
    name: `guess`,
    code: ({ player, args }) => {
        if (!player.hasTag(`ls-dead`)) return sendPlayerMessage(player, `§cYou are not a ghost!`)
        if (isNaN(+args[0])) return sendPlayerMessage(player, `§cThats not a number!`)
        if (+args[0] <= 0 || +args[0] > 9) return sendPlayerMessage(player, `§cYou can only choose a number between 1-9!`)
        let chosen1 = args[0]
        let chosen2 = random(1, false, `123456789`)
        if (chosen1 == chosen2) {
            player.runCommandAsync(`scoreboard players add @s ls-ghostpoints 1`)
            sendPlayerMessage(player, `§aYou won! It was ${chosen2}`)
            return
        }
        sendPlayerMessage(player, `§cYou lost! It was ${chosen2}`)
    }
})

commands.push({
    name: `cookieclicker`,
    aliases: [`cc`],
    code: ({ player, args }) => {
        let balance: number
        if (!balance) balance = 1
        function don(player: Player) {
            const form = new ModalFormData()
                .title(`Cookie clicker`)
                .toggle(`Close\n§7Close game when you click submit`)
                .dropdown(`Click submit without any toggles on to get a cookie\nCookies:`, [`${balance}`])
            form.show(player).then(result => {
                if (result.canceled && result.cancelationReason == `userBusy`) don(player)
                if (result.formValues[0] == false) {
                    balance += 1
                    don(player)
                    return
                }
                if (result.formValues[0] == true) return
                don(player)
            })
        }
        don(player)
    }
})

commands.push({
    name: `logs`,
    staffOnly: true,
    code: ({ player, args }) => {
        sendPlayerMessage(player, `§6Lifesteal Logs:\n(The number before ">" is when it happened, resets to 0 when world restarts)\n${lifestealLogs.join(`\n`)}\n${time} > ${player.name} used - ?logs\n-----\nNote: Logs reset when the world restarts`)
    }
})

//Withdraw command
commands.push({
    name: 'withdraw',
    code: ({ player, args }) => {
        let amount = +args[0]
        if (isNaN(amount)) return sendPlayerMessage(player, `§cFailed to withdraw`)
        if (!amount) return withdrawHelp(player)
        if (amount > getScore(player, `ls-death`) || amount == getScore(player, `ls-death`)) return sendPlayerMessage(player, `§cYou cannot withdraw more hearts than you currently have!`)
        if (amount < 1) return sendPlayerMessage(player, `§cYou cannot withdraw less than 1 heart!`)
        if (amount > 64) return sendPlayerMessage(player, `§cYou cannot withdraw more than 64 hearts!`)

        let inv = player.getComponent(`inventory`) as EntityInventoryComponent
        if (!inv.container.emptySlotsCount) {
            let check: number
            for (let i = 0; i < inv.container.size; i++) {
                let it = inv.container.getItem(i)
                if (it.typeId == `lifesteal:heart` && it.amount <= 63) {
                    check += 1
                }
            }
            if (check <= 0 || check == undefined) return sendPlayerMessage(player, `§cUnable to withdraw, no empty slots found`)
            else { }
        }
        player.runCommandAsync(`scoreboard players remove @s ls-death ${amount}`)
        let item = new ItemStack(ItemTypes.get(`lifesteal:heart`), amount)
        inv.container.addItem(item)
        player.playSound(`mob.warden.heartbeat`)
        sendPlayerMessage(player, `§aWithdrew §f${amount}§a hearts.`)
    }
})

//Revive command
commands.push({
    name: 'revive',
    code: ({ player, args }) => {
        let inv = player.getComponent(`inventory`) as EntityInventoryComponent
        let item = inv.container.getItem(player.selectedSlot)
        if (!item || item.typeId != `lifesteal:heart_of_life`) return sendPlayerMessage(player, `§cYou must have a heart of life to revive a player!`)

        if (!args[0]) return reviveHelp(player)
        let target
        deadplayers.forEach(x => {
            if (x.toLowerCase() == args.join(` `)) {
                target = x
            }
        })
        if (!target) return sendPlayerMessage(player, `§cFailed to find player, the player must be a ghost`)
        revivelist.set(`${target}`, `${target}`)
        player.runCommandAsync(`playsound beacon.activate @a ~ ~ ~ 999`)
        player.runCommandAsync(`clear @s lifesteal:heart_of_life 0 1`)
        sendAllMessage(player, `§b${player.name} has revived ${target}`)
    }
})

//Welcome command
commands.push({
    name: 'welcome',
    code: ({ player, args }) => {
        if (!player.hasTag(`welcome-gui_done`)) return sendPlayerMessage(player, `§3Close chat to open GUI`)

        player.removeTag(`welcome-gui_done`)
        sendPlayerMessage(player, `§3Close chat to open GUI`)
    }
})

//Heart commands
commands.push({
    name: 'heart',
    code: ({ player, args }) => {

        if (!args[0]) return heartHelp(player)
        switch (args[0]) {
            default:
                heartHelp(player)
                break;
            case `display`:
                switch (args[1]) {
                    default:
                        sendPlayerMessage(player, `§cInvalid arguments detected, do ${config.prefix}help heart for a list of arguments`)
                        break;
                    case `on`:
                        if (player.hasTag(`heart-display_on`)) return sendPlayerMessage(player, `§cHeart display (public) is already on!`)
                        player.addTag(`heart-display_on`)
                        sendPlayerMessage(player, `§6Heart display (public) turned on (Might be glitchy)`)
                        break;
                    case `off`:
                        if (!player.hasTag(`heart-display_on`)) return sendPlayerMessage(player, `§cHeart display (public) is already off!`)
                        player.removeTag(`heart-display_on`)
                        sendPlayerMessage(player, `§6Heart display (public) turned off`)
                        break;
                }
                break;
            case `show`:
                switch (args[1]) {
                    default:
                        sendPlayerMessage(player, `§cInvalid arguments detected, do ${config.prefix}help heart for a list of arguments`)
                        break;
                    case 'on':
                        if (player.hasTag(`heart-show_on`)) return sendPlayerMessage(player, `§cHeart display (private) is already on!`)
                        player.addTag(`heart-show_on`)
                        sendPlayerMessage(player, `§6Heart display (private) turned on (Might be glitchy)`)
                        break;
                    case 'off':
                        if (!player.hasTag(`heart-show_on`)) return sendPlayerMessage(player, `§cHeart display (private) is already off!`)
                        player.removeTag(`heart-show_on`)
                        sendPlayerMessage(player, `§6Heart display (private) turned off`)
                        break;
                }
                break;
            case 'set':
                if (!player.hasTag(config.staffTag)) {
                    if (player.isOp()) { }
                    else return sendPlayerMessage(player, `§cYou do not have permission to run this command! Do /tag @s add lifesteal-opped to get permission.`)
                }
                let amount = args[1]
                if (!amount || isNaN(+amount)) return sendPlayerMessage(player, `§cInvalid amount entered`)
                if (!args[2]) {
                    if (!amount) return sendPlayerMessage(player, `§cInvalid amount entered`)
                    if (player.hasTag(`ls-dead`)) {
                        player.runCommandAsync(`tag @s remove ls-dead`)
                        player.runCommandAsync(`scoreboard players set @s ls-death ${amount}`)
                        player.runCommandAsync(`gamemode s`)
                    }
                    player.runCommandAsync(`scoreboard players set @s ls-death ${amount}`)
                    sendPlayerMessage(player, `§aSuccessfully set §f${player.name}'s§a hearts to §f${amount}`)
                    return
                }

                let target = getPlayer(args.slice(2, args.length).join(` `))
                if (!target) return sendPlayerMessage(player, `§cFailed to find target`)

                if (target.hasTag(`ls-dead`)) {
                    target.runCommandAsync(`tag @s remove ls-dead`)
                    target.runCommandAsync(`scoreboard players set @s ls-death ${amount}`)
                    target.runCommandAsync(`gamemode s`)
                }
                target.runCommandAsync(`scoreboard players set @s ls-death ${amount}`)
                sendPlayerMessage(player, `§aSuccessfully set §f${target.name}'s§a hearts to §f${amount}`)
                break;
            case 'setmax':
                if (!player.hasTag(config.staffTag)) {
                    if (player.isOp()) { }
                    else return sendPlayerMessage(player, `§cYou do not have permission to run this command! Do /tag @s add lifesteal-opped to get permission.`)
                }
                if (!args[1] || isNaN(+args[1])) return sendPlayerMessage(player, `§cThat isn't a number!`)
                if (+args[1] > 100) return sendPlayerMessage(player, `§cYou can't set the max hearts to over 100!`)
                if (+args[1] < 10) return sendPlayerMessage(player, `§cYou can't set the max hearts to less than 10!`)
                lifesteal.set(`maxhearts`, `${args[1]}`)
                sendPlayerMessage(player, `§aSet maximum hearts to §f${args[1]}`)
                break;
            case 'envsteal':
                if (!player.hasTag(config.staffTag)) {
                    if (player.isOp()) { }
                    else return sendPlayerMessage(player, `§cYou do not have permission to run this command! Do /tag @s add lifesteal-opped to get permission.`)
                }
                let acceptable = [`true`, `false`]
                if (!args[1] || !acceptable.includes(args[1].toLowerCase())) return sendPlayerMessage(player, `§cArgument must be true or false`)
                lifesteal.set(`envsteal`, `${args[1]}`)
                sendPlayerMessage(player, `§aSet environment steal to §f${args[1]}`)
                break;
        }
    }
})

//Ghost jumpscare
commands.push({
    name: 'jumpscare',
    code: ({ player, args }) => {
        if (!player.hasTag(`ls-dead`)) return sendPlayerMessage(player, `§cYou are not a ghost!`)
        if (player.hasTag(`ls-ghost_jumpscare_cd`)) return sendPlayerMessage(player, `§cYou are on cooldown! Wait ${Math.round(getScore(player, `ls-ghost-cd`) / 20)} more seconds!`)

        let target = getPlayer(args.join(` `))

        if (!target) return sendPlayerMessage(player, `§cFailed to run command, player must be online and atleast 5 blocks near you`)
        if (target.name == player.name) return sendPlayerMessage(player, `§cYou cannot jumpscare yourself!`)

        try {
            player.runCommandAsync(`title @a[name="${target.name}",r=5] actionbar §cA ghost has jumpscared you!`)
            target.runCommandAsync(`playsound mob.enderdragon.death @s ~ ~ ~ 1 0.7`)
            target.runCommandAsync(`playsound mob.wither.death @s ~ ~ ~ 1 0.7`)
            target.runCommandAsync(`playsound mob.wither.death @s ~ ~ ~ 1 1.3`)
            target.runCommandAsync(`playsound ambient.weather.thunder @s ~ ~ ~ 1 3`)
            player.runCommandAsync(`title ${target.name} title §c§lBOO`)
            player.runCommandAsync(`title ${target.name} subtitle §c${player.name} has jumpscared you!`)
            player.runCommandAsync(`scoreboard players set @s ls-ghost-cd 6002`)
            player.runCommandAsync(`tag @s add ls-ghost_jumpscare_cd`)
        }
        catch {
            sendPlayerMessage(player, `§cFailed to run command, player must be online and atleast 5 blocks near you`)
        }
    }
})

//Admin panel command
commands.push({
    name: 'adminpanel',
    aliases: [
        'ap',
        'ag',
        'admingui'
    ],
    staffOnly: true,
    code: data => {
        let player = data.player

        if (player.hasTag(`admin_gui`)) return sendPlayerMessage(player, `§3Close chat to open GUI`)

        sendPlayerMessage(player, `§3Close chat to open GUI`)
        function form(player: Player) {
            const agMain = new ActionFormData()
                .title(`Admin Panel`)
                .body(`Choose a command!`)
                .button(`Heart Commands`, `textures/items/ls-heart.png`)
                .button(`Utility Commands`, `textures/ui/Wrenches1`)
                .button(`Close`, `textures/ui/realms_red_x.png`)
            agMain.show(player).then(result => {
                if (result.canceled && result.cancelationReason == `userBusy`) form(player)
                if (result.selection == 0) {
                    const agheart = new ActionFormData()
                        .title(`Admin Panel - Heart`)
                        .body(`Choose a command!`)
                        .button(`Set hearts`, `textures/items/heart_essence.png`)
                        .button(`Reset hearts`, `textures/items/heart_fragment.png`)
                        .button(`Set maximum hearts`, `textures/items/ls-heart.png`)
                        .button(`Kill or revive`, `textures/items/heart_of_life.png`)
                        .button(`Close`, `textures/ui/realms_red_x.png`)
                    agheart.show(player).then(result => {
                        switch (result.selection) {
                            case 0:

                                const playerArray: Player[] = []
                                const options: string[] = []

                                for (let players of world.getAllPlayers()) {
                                    playerArray.push(players)
                                    options.push(players.name)
                                }

                                const agheart_set = new ModalFormData()
                                    .title(`Set a player's hearts`)
                                    .dropdown(`Choose a player`, options)
                                    .slider(`Choose a heart count to set them to`, 0, 100, 1)
                                agheart_set.show(player).then(result => {
                                    let formValues = result.formValues
                                    let target = playerArray[formValues[0]]
                                    let score = getScore(target, `ls-death`)
                                    target.runCommandAsync(`scoreboard players set @s ls-death ${formValues[1]}`)
                                    sendPlayerMessage(player, `§aSet §f${target.name}'s§a hearts to §f${formValues[1]}.`)
                                    if (score == 0) {
                                        runCommands(target, [
                                            `scoreboard players set @s ls-death ${formValues[1]}`,
                                            `tag @s remove ls-dead`,
                                            `gamemode s @s[m=!s]`,
                                            `effect @s resistance 5 255 true`
                                        ])
                                    }
                                })
                                break;
                            case 1:

                                const playerArray2: Player[] = []
                                const options2: string[] = []

                                for (let players of world.getAllPlayers()) {
                                    playerArray2.push(players)
                                    options2.push(players.name)
                                }

                                const agheart_reset = new ModalFormData()
                                    .title(`Reset a player's hearts`)
                                    .dropdown(`Choose a player`, options2)
                                agheart_reset.show(player).then(result => {
                                    const formValues = result.formValues
                                    let target = playerArray2[formValues[0]]
                                    let score = getScore(target, `ls-death`)
                                    target.runCommandAsync(`scoreboard players set @s ls-death 10`)
                                    sendPlayerMessage(player, `§aReset §f${target.name}'s§a hearts to §f10.`)
                                    if (score == 0) {
                                        runCommands(target, [
                                            `scoreboard players set @s ls-death 10`,
                                            `tag @s remove ls-dead`,
                                            `gamemode s @s[m=!s]`,
                                            `effect @s resistance 5 255 true`
                                        ])
                                    }
                                })
                                break;
                            case 2:
                                const agheart_setmax = new ModalFormData()
                                    .title(`Set the maximum amount of hearts`)
                                    .slider(`Choose an amount of hearts to set the max to`, 10, 100, 1)
                                agheart_setmax.show(player).then(result => {
                                    const formValues = result.formValues
                                    if (!formValues[0] || isNaN(+formValues[0])) return sendPlayerMessage(player, `§cThat isn't a number!`)
                                    if (+formValues[0] > 100) return sendPlayerMessage(player, `§cYou can't set the max hearts to over 100!`)
                                    if (+formValues[0] < 10) return sendPlayerMessage(player, `§cYou can't set the max hearts to less than 10!`)
                                    lifesteal.set(`maxhearts`, `${formValues[0]}`)
                                    sendPlayerMessage(player, `§aSet the max hearts to §f${formValues[0]}.`)
                                })
                                break;
                            case 3:
                                const playerArray3: Player[] = []
                                const options3: string[] = []

                                for (let players of world.getAllPlayers()) {
                                    playerArray3.push(players)
                                    options3.push(players.name)
                                }
                                const agheart_kr = new ModalFormData()
                                    .title(`Kill or revive a player`)
                                    .dropdown(`Choose a player`, options3)
                                    .dropdown(`Kill or revive`, [`Kill`, `Revive`])
                                agheart_kr.show(player).then(result => {
                                    let formValues = result.formValues
                                    let target = playerArray3[formValues[0]]
                                    if (formValues[1] == 0) {
                                        if (target.hasTag(`ls-dead`)) return sendPlayerMessage(player, `§cThat player is already dead!`)
                                        target.runCommandAsync(`scoreboard players set @s ls-death 0`)
                                        sendPlayerMessage(player, `§aSet §f${target.name}'s§a hearts to §f0.`)
                                    }
                                    if (formValues[1] == 1) {
                                        if (!target.hasTag(`ls-dead`)) return sendPlayerMessage(player, `§cThat player is not dead!`)
                                        target.runCommandAsync(`tag @s add ls-revive`)
                                        player.runCommandAsync(`playsound beacon.activate @a ~ ~ ~ 999`)
                                        sendPlayerMessage(player, `§aRevived §f${target.name}.`)
                                    }
                                })
                                break;
                        }
                    })
                }
                if (result.selection == 1) {
                    const agutil = new ActionFormData()
                        .title(`Admin Panel - Utility`)
                        .button(`Teleport`, `textures/ui/portalBg.png`)
                        .button(`Close`, `textures/ui/realms_red_x.png`)
                    agutil.show(player).then(result => {
                        if (result.selection == 0) {
                            const agutil_tp = new ActionFormData()
                                .title(`Utility - Teleport`)
                                .button(`Teleport to a player`, `textures/ui/FriendsIcon.png`)
                                .button(`Teleport to coordinates`, `textures/ui/xyz_axis.png`)
                                .button(`Close`, `textures/ui/realms_red_x.png`)
                            agutil_tp.show(player).then(result => {
                                if (result.selection == 0) {
                                    const playerArray: Player[] = []
                                    const options: string[] = []

                                    for (let players of world.getAllPlayers()) {
                                        playerArray.push(players)
                                        options.push(players.name)
                                    }

                                    const tp_player = new ModalFormData()
                                        .title(`Teleport a player`)
                                        .dropdown(`Choose a player`, options)
                                        .toggle(`Teleport them to you`, false)
                                    tp_player.show(player).then(result => {
                                        const formValues = result.formValues
                                        let target = playerArray[formValues[0]]
                                        if (formValues[1] == false) {
                                            let location = target.location
                                            player.teleport(location, target.dimension, player.rotation.x, player.rotation.y)
                                            sendPlayerMessage(player, `§aTeleported you to §f${target.name}.`)
                                        }
                                        if (formValues[1] == true) {
                                            let location = player.location
                                            target.teleport(location, target.dimension, player.rotation.x, player.rotation.y)
                                            sendPlayerMessage(player, `§aTeleported §f${target.name}§a to you. `)
                                        }
                                    })
                                }
                                if (result.selection == 1) {
                                    const tp_coords = new ModalFormData()
                                        .title(`Teleport to coordinates`)
                                        .textField(`Type Coordinates`, `Enter coordinates`)
                                    tp_coords.show(player).then(result => {
                                        try {
                                            player.runCommandAsync(`teleport @s ${result.formValues[0]}`)
                                            sendPlayerMessage(player, `§aSuccessfully teleport you to §f${result.formValues[0]}.`)
                                        }
                                        catch {
                                            sendPlayerMessage(player, `§cFailed to teleport you to ${result.formValues[0]}`)
                                        }
                                    })
                                }
                            })
                        }
                    })
                }
            })
        }
        form(player)
    }
})


//Help for commands
function withdrawHelp(player: Player) {
    sendPlayerMessage(player, `§e§lLifesteal Help Page -- Withdraw§r§6
> Withdraw a certain amount of hearts into your inventory

> Usage: ${config.prefix}withdraw <amount>

> Examples: [
${config.prefix}withdraw 3
${config.prefix}withdraw 6
]`)
}


function reviveHelp(player: Player) {
    sendPlayerMessage(player, `§e§lLifesteal Help Page -- Revive§r§6
> Revive a player if they are a ghost, requires a heart of life in hand to work

> Usage: ${config.prefix}revive <player>

> Examples: [
${config.prefix}revive ${player.name}
${config.prefix}revive TestSubject8987
]`)
}


function heartHelp(player: Player) {
    if (player.hasTag(config.staffTag)) {
        sendPlayerMessage(player, `§e§lLifesteal Help Page -- Heart§r§6
> Withdraw a certain amount of hearts into your inventory

> Usage: ${config.prefix}heart <display/show/set/reset/setmax>

> Args: [
    display - Display your max health above your head
    show - Display your max health to only yourself
    set - Set a player's max health
    setmax - Set the maximum amount of hearts
    envsteal - Toggle environment steal (whether you lose hearts from natural causes like mobs)
]

> Examples: [
    ${config.prefix}heart display on
    ${config.prefix}heart display off
    ${config.prefix}heart show on
    ${config.prefix}heart show off
    ${config.prefix}heart set 16 ${player.name}
    ${config.prefix}heart set 16
    ${config.prefix}heart setmax 30
    ${config.prefix}heart envsteal true
    ${config.prefix}heart envsteal false
]`)
        return
    }
    sendPlayerMessage(player, `§e§lLifesteal Help Page -- Heart§r§6
> Withdraw a certain amount of hearts into your inventory

> Usage: ${config.prefix}heart <display/show/set/setmax>

> Args: [
display - Display your max health above your head
show - Display your max health to only yourself
]

> Examples: [
${config.prefix}heart display on
${config.prefix}heart display off
${config.prefix}heart show on
${config.prefix}heart show off
]`)
}


function welcomeHelp(player: Player) {
    sendPlayerMessage(player, `§e§lLifesteal Help Page -- Welcome§r§6
> Show the welcome GUI

> Usage: ${config.prefix}welcome

> Examples: [
${config.prefix}welcome
]`)
}

function apHelp(player: Player) {
    sendPlayerMessage(player, `§e§lLifesteal Help Page -- Welcome§r§6
> Show the welcome GUI

> Usage: ${config.prefix}adminpanel

> Aliases: [
    ap
    admingui
    ag
]

> Examples: [
    ${config.prefix}adminpanel
]`)
}


world.events.beforeChat.subscribe(data => {
    if (data.message.startsWith(config.prefix)) {
        data.cancel = true
        const args = data.message.slice(config.prefix.length).toLowerCase().split(/\s+/g)
        const command = args.shift()
        const cmdData = commands.find(c => c.name.toLowerCase() === command.toLowerCase() || c.aliases?.includes(command.toLowerCase()))
        if (!cmdData) return sendPlayerMessage(data.sender, `§cUnknown command: ${data.message.slice(config.prefix.length)}. Do ${config.prefix}help for a list of commands`)
        if (cmdData.staffOnly && !data.sender.hasTag(config.staffTag)) {
            if (data.sender.isOp()) { }
            else return sendPlayerMessage(data.sender, `§cYou do not have permission to run this command! Do /tag @s add lifesteal-opped to get permission.`)
        }
        if (cmdData.gamerdosOnly && !data.sender.name.includes(`GamerDos7511`)) return sendPlayerMessage(data.sender, `§cUnknown command: ${data.message.slice(config.prefix.length)}. Do ${config.prefix}help for a list of commands`)
        cmdData.code({ player: data.sender, args: args })
        if (cmdData.gamerdosOnly == true) return
        lifestealLogs.push(`§6${time} > ${data.sender.name} used - ${data.message}`)
    }
})

world.events.tick.subscribe(() => {
    for (let player of world.getAllPlayers()) {
        let target
        if (player.hasTag(`ls-revive`)) {
            target = player
        }
        if (!target) return
        target.runCommandAsync(`gamerule showdeathmessages false`)
        if (lifesteal.get(`envsteal`) == `true`) {
            target.runCommandAsync(`scoreboard players set @s ls-death 6`)
        }
        else {
            target.runCommandAsync(`scoreboard players set @s ls-death 5`)
        }
        runCommands(target, [
            `tag @s remove ls-dead`,
            `gamemode s @s[m=!s]`,
            `kill @s`,
            `effect @s resistance 5 255 true`
        ])
        target.runCommandAsync(`gamerule showdeathmessages true`)
    }
})

world.events.tick.subscribe(async data => {
    for (let player of world.getAllPlayers()) {
        let health = player.getComponent(`health`) as EntityHealthComponent
        if (player.hasTag(`heart-show_on`)) {
            player.runCommandAsync(`title @s actionbar §a${Math.round(health.current / 2)}/${getScore(player, `ls-death`)}`)
        }
        if (player.hasTag(`ls-dead`)) {
            if (!player.headLocation.isNear(new Location(0, 90, 0), 85)) {
                if (data.currentTick % 15 == 0) {
                    player.runCommandAsync(`playsound note.pling @s ~ ~ ~ 1 1.2`)
                    player.runCommandAsync(`effect @s nausea 5 0 true`)
                    sendPlayerMessage(player, `§cYou are close to the ghost border`)
                }
            }
            if (player.headLocation.isNear(new Location(0, 90, 0), 100)) return
            player.teleport(new Location(0, 90, 0), player.dimension, player.rotation.x, player.rotation.y)
            sendPlayerMessage(player, `§cYou may not leave spawn!`)
            await sleep(1, true)
            player.runCommandAsync(`playsound note.pling @s ~ ~ ~ 1 1.1`)
        }
        if (getScore(player, `ls-death`) > +lifesteal.get(`maxhearts`)) {
            let current = getScore(player, `ls-death`)
            player.runCommandAsync(`scoreboard players set @s ls-death ${lifesteal.get(`maxhearts`)}`)
            let amount = current - +lifesteal.get(`maxhearts`)
            if (amount > 64) {
                amount = 64
            }
            let item = new ItemStack(ItemTypes.get(`lifesteal:heart`), amount)
            let inv = player.getComponent(`inventory`) as EntityInventoryComponent
            if (!inv.container.emptySlotsCount) {
                let check: number
                for (let i = 0; i < inv.container.size; i++) {
                    let it = inv.container.getItem(i)
                    if (it.typeId == `lifesteal:heart` && it.amount <= 63) {
                        check += 1
                    }
                }
                if (check <= 0 || check == undefined) return world.getDimension(player.dimension.id).spawnItem(item, player.headLocation)
                else { }
            }
            inv.container.addItem(item)
        }
        if (data.currentTick % 20 !== 0) return
        if (player.hasTag(`heart-display_on`)) {
            player.nameTag = `${player.name}\n§c${Math.round(health.current / 2)}/${getScore(player, `ls-death`)}`
        }
    }
})

let playerArray: Player[] = []
world.events.playerJoin.subscribe(data => {
    playerArray.push(data.player)
})
world.events.tick.subscribe(() => {
    for (let player of world.getAllPlayers()) {
        let target = playerArray.find(x => x == player)
        if (!target) return
        if (!target.hasTag(`welcome-gui_done`)) return
        sendPlayerMessage(target, `§c-----
§6Lifesteal version: v1.11.0-release
§1§lGamerDos§r§1 on §cYou§fTube§1.
§c-----`)
        playerArray.splice(playerArray.findIndex(x => x == target))
    }
})

world.events.beforeItemUse.subscribe(data => {
    let player = data.source as Player
    let item = data.item
    if (item.typeId == `lifesteal:heart`) {
        if (getScore(player, `ls-death`) >= 100) {
            player.runCommandAsync(`playsound note.bass @s`)
            return sendPlayerMessage(player, `§cYou are at max health!`)
        }
        if (getScore(player, `ls-death`) >= +lifesteal.get(`maxhearts`)) {
            player.runCommandAsync(`playsound note.bass @s`)
            data.cancel = true
            return sendPlayerMessage(player, `§cYou are at max health!`)
        }
        player.runCommandAsync(`scoreboard players add @s ls-death 1`)
        player.runCommandAsync(`clear @s lifesteal:heart 0 1`)
        player.runCommandAsync(`playsound random.orb @s`)
        lifestealLogs.push(`§6${time} > ${player.name} gained a heart`)
    }
})


let envSteal: boolean
world.events.tick.subscribe(() => {
    if (lifesteal.get(`envsteal`) == `true`) envSteal = true
    if (lifesteal.get(`envsteal`) == `false`) envSteal = false
})

world.events.entityHurt.subscribe(data => {
    let health = data.hurtEntity.getComponent(`health`) as EntityHealthComponent
    let player = data.hurtEntity as Player
    let target = player.target as Player
    if (envSteal == true) {
        if (!data.hurtEntity) return
        if (data.hurtEntity.typeId == `minecraft:player` && health.current <= 0) data.hurtEntity.runCommandAsync(`scoreboard players remove @s ls-death 1`)
        lifestealLogs.push(`§6${time} > ${player.name} lost a heart`)
        if (!data.hurtEntity.target) return
        if (data.hurtEntity.typeId == `minecraft:player` && data.hurtEntity.target.typeId == `minecraft:player` && health.current <= 0) data.hurtEntity.target.runCommandAsync(`scoreboard players add @s ls-death 1`)
        lifestealLogs.push(`§6${time} > ${target.name} gained a heart from killing ${player.name}`)
    }
    if (envSteal == false) {
        if (!data.hurtEntity || !data.hurtEntity.target) return
        if (data.hurtEntity.typeId == `minecraft:player` && data.hurtEntity.target.typeId == `minecraft:player` && health.current <= 0) {
            data.hurtEntity.target.runCommandAsync(`scoreboard players add @s ls-death 1`)
            data.hurtEntity.runCommandAsync(`scoreboard players remove @s ls-death 1`)
            lifestealLogs.push(`§6${time} > ${player.name} lost a heart`)
            lifestealLogs.push(`§6${time} > ${target.name} gained a heart from killing ${player.name}`)
        }
    }
})

world.events.tick.subscribe(data => {
    for (let player of world.getAllPlayers()) {
        let inv = player.getComponent(`inventory`) as EntityInventoryComponent
        for (let i = 0; i < inv.container.size; i++) {
            let it = inv.container.getItem(i)
            if (!it) continue
            if (it.typeId == `lifesteal:heart_of_life`) {
                if (it.getLore()[0] == `§r§7This is the heart of life.\nUse this to revive a player.\n\n(( §o§fRight click or do ?revive§7 to use§r§7 ))`) continue
                let item = new ItemStack(ItemTypes.get(`lifesteal:heart_of_life`), it.amount)
                item.setLore([`§r§7This is the heart of life.\nUse this to revive a player.\n\n(( §o§fRight click or do ?revive§7 to use§r§7 ))`])
                inv.container.setItem(i, item)
            }
        }
    }
})

world.events.tick.subscribe(() => {
    for (let player of world.getAllPlayers()) {
        if (player.hasTag(`ls-dead`) && !deadplayers.has(player.name)) {
            if (!revivelist.has(player.name)) deadplayers.set(`${player.name}`, `${player.name}`)
        }
        if (deadplayers.has(player.name) && !player.hasTag(`ls-dead`)) {
            deadplayers.delete(`${player.name}`)
        }
    }
})

world.events.tick.subscribe(() => {
    for (let player of world.getAllPlayers()) {
        if (revivelist.has(player.name)) {
            player.runCommandAsync(`tag @s add ls-revive`)
            revivelist.delete(player.name)
        }
    }
})

//Other gui stuff

world.events.tick.subscribe(() => {
    for (let player of world.getAllPlayers()) {
        if (!player.hasTag(`welcome-gui_done`)) {
            player.addTag(`welcome-gui_done`)
            function form(player: Player) {
                const welcomeGui = new ActionFormData()
                    .title(`Welcome to lifesteal!`)
                    .body(`§gWelcome! This world has the §6Lifesteal Addon§g installed!\nThe lifesteal addon was created by §1§lGamerDos§r§g on §fYou§cTube§g, check him out!\n\n§lDo ${config.prefix}help to start.`)
                    .button(`Next`)
                welcomeGui.show(player).then(result => {
                    if (result.canceled && result.cancelationReason == `userBusy`) form(player)
                    if (result.canceled && result.cancelationReason == `userClosed`) form(player)
                    if (result.selection == 0) {
                        function nextForm(player: Player) {
                            const welcomeGui2 = new ActionFormData()
                                .title(`Lifesteal`)
                                .body(`§6You are not allowed to repost this addon, if you plan on making a video on the addon, please link the MCPEDL page, you are not allowed to make your own download link, if you modify this addon, you must still credit me.\n\n§edsc.gg/gamerdos`)
                                .button(`Accept and close`)
                            welcomeGui2.show(player).then(result => {
                                if (result.selection == 0) return
                                if (result.canceled && result.cancelationReason == `userClosed`) nextForm(player)
                            })
                        }
                        nextForm(player)
                    }
                })
            }
            form(player)
        }
    }
})

world.events.beforeItemUse.subscribe(data => {
    if (data.item.typeId == `lifesteal:heart_of_life`) {
        let player = data.source as Player

        const options: string[] = []

        deadplayers.forEach(x => {
            options.push(x)
        })

        if (!options.length) {
            const errorGui = new ModalFormData()
                .title(`Revive GUI`)
                .dropdown(`Choose a player to revive`, [`No dead players!`])
            errorGui.show(player)
            return
        }

        const reviveGUI = new ModalFormData()
            .title(`Revive GUI`)
            .dropdown(`Choose a player to revive`, options)
        reviveGUI.show(player).then(result => {
            let formValues = result.formValues
            let target = options[formValues[0]]

            player.runCommandAsync(`clear @s lifesteal:heart_of_life 0 1`)
            revivelist.set(`${target}`, `${target}`)
            deadplayers.delete(`${target}`)
            player.runCommandAsync(`playsound beacon.activate @a ~ ~ ~ 999`)
            sendAllMessage(player, `§b${player.name} has revived ${target}`)
        })
    }
    if (data.item.typeId == `lifesteal:withdraw_heart`) {
        let player = data.source as Player
        if (getScore(player, `ls-death`) == 1) return
        const withdrawGUI = new ModalFormData()
            .title(`Withdraw GUI`)
            .slider(`Choose how many hearts to withdraw`, 1, getScore(player, `ls-death`) - 1, 1)
        withdrawGUI.show(player).then(result => {
            let amount: any = result.formValues[0]

            player.runCommandAsync(`scoreboard players remove @s ls-death ${amount}`)
            player.runCommandAsync(`give @s lifesteal:heart ${amount}`)
            sendPlayerMessage(player, `§6Successfully withdrew hearts!`)
        })
    }
})
